import psycopg2
from django.apps import apps
from django.core.management.base import BaseCommand

from urlid_graph import settings as urlid_graph_settings
from urlid_graph.log import RemoveDuplicateObjectsLogger
from urlid_graph.utils import remove_duplicates_sql


class Command(BaseCommand):
    help = "Remove duplicates for an object table"

    def add_arguments(self, parser):
        parser.add_argument("--disable-autovacuum", action="store_true")
        parser.add_argument("--disable-indexes", action="store_true")
        parser.add_argument("app_name")
        parser.add_argument("model")

    def handle(self, *args, **options):
        disable_autovacuum = options["disable_autovacuum"]
        disable_indexes = options["disable_indexes"]
        app_name = options["app_name"]
        model = options["model"]
        Model = apps.get_model(app_name, model)

        self.logger = RemoveDuplicateObjectsLogger(description=f"{app_name}.{model}")
        print(f"Starting job {self.logger.job.id}")

        connection = psycopg2.connect(urlid_graph_settings.DJANGO_DATABASE_URL)
        connection.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        with connection.cursor() as cursor:
            self.logger.start_step("execute-query")
            queries = remove_duplicates_sql(
                table_name=Model._meta.db_table,
                field_names=[field.name for field in Model._meta.fields if field.name != "id"],
                disable_autovacuum=disable_autovacuum,
                disable_indexes=disable_indexes,
                disable_sync_commit=True,
                disable_triggers=True,
            )
            for counter, query in enumerate(queries, start=1):
                self.logger.progress(done=counter, total=len(queries))
                cursor.execute(query)
            self.logger.finish()



# Generated by Django 4.0.6 on 2022-08-02 18:31

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('urlid_graph', '0004_alter_joblog_id_alter_logstep_message'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='joblog',
            options={'ordering': ('-created_at',)},
        ),
        migrations.AlterModelOptions(
            name='logstep',
            options={'ordering': ('-created_at',)},
        ),
    ]



from contextlib import contextmanager
from dataclasses import dataclass

from urlid_graph.models import JobLog, LogStep


@dataclass
class Step:
    slug: str
    description: str


class AsyncLogger:
    name: str = None
    steps: list = None

    def __init__(self, description):
        self.description = description
        self.job = self._init_job()
        self._current_step = None

        if self.name is None:
            raise ValueError("Attribute 'name' cannot be None")

        if self.steps is None:
            raise ValueError("Attribute 'steps' cannot be None")

        self._step_slugs = [s.slug for s in self.steps]

    def _init_job(self):
        return JobLog.objects.create(name=self.name, description=self.description)

    def start_step(self, step, message=None):
        if step not in self._step_slugs:
            raise ValueError(f"Step '{step}' not found")

        if self._current_step is not None:
            self.finish_step(self._current_step, force=True)

        LogStep.objects.create(job=self.job, step=step, action="start-step", message=message)
        self._current_step = step

    def progress(self, done, total, message=None):
        if self._current_step is None:
            raise ValueError(f"Cannot make progress in a job which current step is None")

        LogStep.objects.create(
            job=self.job, step=self._current_step, action="progress", done=done, total=total, message=message
        )

    def finish_step(self, step, message=None, force=False):
        if step not in self._step_slugs:
            raise ValueError(f"Step '{step}' not found")

        if not force and (self._current_step is None or self._current_step != step):
            raise ValueError(f"Current step different from '{self._current_step}'")

        LogStep.objects.create(job=self.job, step=step, action="finish-step", message=message)
        self._current_step = None

    def message(self, message):
        current_step = self._current_step or "no-step"
        LogStep.objects.create(job=self.job, step=current_step, action="message", message=message)

    def error(self, message):
        current_step = self._current_step or "no-step"
        LogStep.objects.create(job=self.job, step=current_step, action="error", message=message)

    def finish(self, message=None):
        if self._current_step is not None:
            self.finish_step(self._current_step, force=True)

        LogStep.objects.create(job=self.job, step="complete", action="finish", message=message)

    def __str__(self):
        return f"{self.name} - {self.description}: {self.job.job_id}"

    @contextmanager
    def step(self, step, message_start=None, message_finish=None):
        self.start_step(step, message=message_start)
        yield
        self.finish_step(step, message=message_finish)


class ImportObjectsLogger(AsyncLogger):
    name = "importing objects"
    steps = [
        Step("pre-import", "Disable indexes etc."),
        Step("import", "Import data into table"),
        Step("index", "Index imported objects in ObjectRepository"),
        Step("post-import", "Re-enable indexes etc."),
    ]


class ImportRelationshipsLogger(AsyncLogger):
    name = "importing relationships"
    steps = [
        Step("pre-import", "Disable indexes etc."),
        Step("import-table-data", "Import data into table"),
        Step("import-relationships", "Import data into table"),
        Step("post-import", "Re-enable indexes etc."),
    ]


class IndexObjectsLogger(AsyncLogger):
    name = "indexing objects"
    steps = [
        Step("delete-objects", "Delete already indexed objects for this model"),
        Step("index", "Index objects in ObjectRepository"),
    ]


class RemoveDuplicateObjectsLogger(AsyncLogger):
    name = "removing duplicate objects"
    steps = [
        Step("execute-query", "Execute N queries to remove duplicates/insert"),
    ]



import agensgraph  # noqa
from django.conf import settings

USER_SETTINGS = getattr(settings, "URLID_GRAPH", dict())

DEFAULTS = {
    "SEARCH_LANGUAGE": "pg_catalog.portuguese",
    "RELATIONSHIP_DEPTH_LIMIT": 3,
    "ALL_NODES_RELATIONSHIPS_CHUNK_SIZE": 10,
    "GRAPH_NAME": "graph_db",
    "NODES_CSV_SIZE_LIMIT": 1000,
}


def get_settings(key):
    return USER_SETTINGS.get(key, DEFAULTS[key])


SEARCH_LANGUAGE = get_settings("SEARCH_LANGUAGE")
RELATIONSHIP_DEPTH_LIMIT = get_settings("RELATIONSHIP_DEPTH_LIMIT")
ALL_NODES_RELATIONSHIPS_CHUNK_SIZE = get_settings("ALL_NODES_RELATIONSHIPS_CHUNK_SIZE")
GRAPH_NAME = get_settings("GRAPH_NAME")
NODES_CSV_SIZE_LIMIT = get_settings("NODES_CSV_SIZE_LIMIT")

DJANGO_DATABASE_URL = settings.DATABASE_URL
DJANGO_DATABASE = "default"
GRAPH_DATABASE = getattr(settings, "GRAPH_DATABASE", None)
GRAPH_DATABASE_URL = getattr(settings, "GRAPH_DATABASE_URL", None)

TESTING_ENV = getattr(settings, "TESTING_ENV", False)



# Generated by Django 4.0.6 on 2022-07-28 14:19
import uuid

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('urlid_graph', '0002_alter_elementconfig_options_alter_entity_options_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='JobLog',
            fields=[
                ('id', models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)),
                ('name', models.CharField(max_length=255)),
                ('description', models.CharField(max_length=255)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='LogStep',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('action', models.CharField(max_length=255)),
                ('step', models.CharField(max_length=255)),
                ('done', models.FloatField(blank=True, null=True)),
                ('total', models.FloatField(blank=True, null=True)),
                ('message', models.CharField(blank=True, max_length=255, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('job', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='steps', to='urlid_graph.joblog')),
            ],
        ),
    ]



from django.contrib import admin
from django.urls import reverse
from django.utils import timezone
from django.utils.html import format_html

from urlid_graph.models import ElementConfig, Entity, JobLog, LogStep, ObjectRepository, SavedGraph


class JobLogAdmin(admin.ModelAdmin):
    list_display = (
        "id", "name", "description", "current", "progress", "eta",
        "duration", "created", "updated",
    )
    search_fields = ("id", "name", "description")
    readonly_fields = ("progress", "eta", "current_step", "last_updated_at")
    list_filter = ("name",)

    def current(self, instance):
        url = reverse(f"admin:urlid_graph_logstep_add").replace("add/", "") + f"?job__id={instance.id}"
        return format_html(f'{instance.current_step} (<a href="{url}">list</a>)')
    current.description = "Current step"

    def created(self, instance):
        return instance.created_at.strftime("%Y-%m-%d %H:%M:%S %Z")

    def updated(self, instance):
        return instance.created_at.strftime("%Y-%m-%d %H:%M:%S %Z")

    def duration(self, instance):
        if instance.last_updated_at is None:
            return None
        if instance.current_step == "complete":
            return instance.last_updated_at - instance.created_at
        else:
            return timezone.now() - instance.created_at


class LogStepAdmin(admin.ModelAdmin):
    list_display = ("id", "job_id", "created_at", "action", "step", "message", "done", "total")
    search_fields = ("job__id", "action", "step", "message")
    list_filter = ("action", "step")

    def job_id(self, instance):
        if instance.id:
            return instance.job.id
        return "-"

    job_id.short_description = "Job ID"


class EntityAdmin(admin.ModelAdmin):
    pass


class ElementConfigAdmin(admin.ModelAdmin):
    pass


class ObjectRepositoryAdmin(admin.ModelAdmin):
    pass


class SavedGraphAdmin(admin.ModelAdmin):
    pass


admin.site.register(JobLog, JobLogAdmin)
admin.site.register(LogStep, LogStepAdmin)
admin.site.register(Entity, EntityAdmin)
admin.site.register(ElementConfig, ElementConfigAdmin)
admin.site.register(ObjectRepository, ObjectRepositoryAdmin)
admin.site.register(SavedGraph, SavedGraphAdmin)



class UnexpectedCypherQueryTypeException(Exception):
    """
    Raised when the query type (node, path or edge) is not the expected one
    """


class InvalidEdgeIdException(Exception):
    """
    Raised for invalid edge id formats
    """


class InvalidRelationshipsDirectionsException(Exception):
    """
    Raised when fetching for relationships using invalid inbound/outbound directions
    """



import datetime
import pprint
import random
import string
import textwrap
import time
from contextlib import contextmanager
from itertools import zip_longest
from textwrap import dedent

import psycopg2
from rows.utils import uncompressed_size


def random_name(n=16):
    return "".join(random.choice(string.ascii_letters) for _ in range(n))


@contextmanager
def working(message, finish="done in {duration}."):
    print(f"{message}... ", flush=True, end="")
    start = time.time()
    yield
    end = time.time()
    print(finish.format(duration=datetime.timedelta(seconds=end - start)))


def read_total_size(source):
    try:
        total_size = uncompressed_size(source)
    except (RuntimeError, ValueError):
        total_size = None

    return total_size


class DatabaseConnection:

    # TODO: move `working` logic to a logger with different levels

    def __init__(self, uri=None, connection=None, debug=False):
        if (not uri and not connection) or (uri and connection):
            raise ValueError("Only one parameter (uri or connection) must be specified")
        self._connection = connection
        self._uri = uri
        self.debug = debug

    @property
    def connection(self):
        if not self._connection:
            self._connection = psycopg2.connect(self._uri)

        return self._connection

    def debug_message(self, query, params):
        if not isinstance(query, str):  # psycopg2.sql.SQL instance
            query = query.as_string(self.connection)
        output = []
        for line in textwrap.dedent(query).strip().splitlines():
            output.extend(textwrap.wrap(line, 80))
        query_text = textwrap.indent("\n".join(output), "*  ")

        message = f"\n* Executing query:\n{query_text}\n"
        if params is not None:
            params_text = textwrap.indent(pprint.pformat(params), "*  ")
            message += f"* with params:\n{params_text}\n"
        return message

    def execute_query(self, query, params=None):
        with self.connection.cursor() as cursor:
            if self.debug:
                with working(self.debug_message(query, params)):
                    cursor.execute(query, params)
            else:
                cursor.execute(query, params)
            return cursor

    def execute_queries(self, queries, params_list=None):
        params_list = params_list or []
        with self.connection.cursor() as cursor:
            for query, params in zip_longest(queries, params_list):
                if self.debug:
                    with working(self.debug_message(query, params)):
                        cursor.execute(query, params)
                else:
                    cursor.execute(query, params)
            return cursor

    def enable_trigger(self, table_name, trigger):
        return self.execute_query(f"ALTER TABLE {table_name} ENABLE TRIGGER {trigger}")

    def disable_trigger(self, table_name, trigger):
        return self.execute_query(f"ALTER TABLE {table_name} DISABLE TRIGGER {trigger}")

    def disable_sync_commit(self):
        return self.execute_query("SET synchronous_commit = off")

    def enable_sync_commit(self):
        return self.execute_query("SET synchronous_commit = on")

    def enable_autovacuum(self, table_name=None):
        if table_name is not None:
            queries = [f"ALTER TABLE {table_name} SET (autovacuum_enabled = true)"]
        else:  # Enable for the whole database
            queries = [
                "ALTER SYSTEM SET autovacuum = on",
                "SELECT pg_reload_conf()",
            ]
        return self.execute_queries(queries)

    def disable_autovacuum(self, table_name=None):
        if table_name is not None:
            queries = [f"ALTER TABLE {table_name} SET (autovacuum_enabled = false)"]
        else:  # Disable for the whole database
            queries = [
                "ALTER SYSTEM SET autovacuum = off",
                "SELECT pg_reload_conf()",
            ]
        return self.execute_queries(queries)

    def vacuum_analyze(self, table_name=None):
        return self.execute_query(f"VACUUM ANALYZE {table_name or ''}")

    def delete_table(self, table_name):
        self.execute_query(f"DROP TABLE {table_name}")


def remove_duplicates_sql(
    table_name,
    field_names,
    disable_autovacuum=False,
    disable_indexes=False,
    disable_sync_commit=True,
    disable_triggers=True,
):

    before, queries, after = [], [], []
    field_list = ", ".join(field_names)

    if disable_autovacuum:
        before.append(f"ALTER TABLE {table_name} SET (autovacuum_enabled = false);")

    queries.extend(
        [
            "BEGIN;",
            f"""
            CREATE TEMP TABLE {table_name}_unique ON COMMIT DROP AS
                SELECT DISTINCT
                    {field_list}
                FROM {table_name};
            """,
            f"""
            TRUNCATE TABLE {table_name};
            """,
        ]
    )
    if disable_sync_commit:
        queries.append("SET LOCAL synchronous_commit = off;")
    if disable_indexes:
        queries.append(f"UPDATE pg_index SET indisready = FALSE WHERE indrelid = '{table_name}'::regclass;")
    if disable_triggers:
        queries.append(f"ALTER TABLE {table_name} DISABLE TRIGGER ALL;")
    queries.append(
        f"""
        INSERT INTO {table_name} ({field_list})
            SELECT {field_list} FROM {table_name}_unique;
        """
    )
    if disable_indexes:
        queries.append(f"REINDEX TABLE {table_name};")
    if disable_triggers:
        queries.append(f"ALTER TABLE {table_name} ENABLE TRIGGER ALL;")
    if disable_sync_commit:
        queries.append("RESET synchronous_commit;")
    queries.append("END;")
    after.append(f"VACUUM ANALYZE {table_name};")

    if disable_autovacuum:
        queries.append(f"ALTER TABLE {table_name} RESET (autovacuum_enabled);")

    return [
        query
        for query in (
            "\n".join(dedent(query) for query in before),
            "\n".join(dedent(query) for query in queries),
            "\n".join(dedent(query) for query in after),
        )
        if query.strip()
    ]



import csv
from collections import defaultdict

from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404
from rest_framework import viewsets
from rest_framework.response import Response
from rest_framework.views import APIView

from .exceptions import InvalidEdgeIdException
from .mixins import GraphDbMixin
from .models import DatasetModel, Entity, ObjectRepository, SavedGraph, get_nodes_and_properties
from .serializers import (
    AllNodesRelationshipsSerializer,
    ConfigSerializer,
    DetailedEdgeSerializer,
    DetailedNodeSerializer,
    RelationshipArgumentsSerializer,
    SavedGraphSerializer,
    UUIDListSerializer,
)


class GraphDataVisConfig(APIView):
    # TODO this endpoint should extends a base urlid_views one
    def get(self, request):
        serializer = ConfigSerializer({"entities": Entity.objects.all()})
        return Response(serializer.data)


class ExportVerticesCSVView(APIView):
    # TODO this endpoint should extends a base urlid_views one
    def build_csv(self, objs):
        all_properties, header_per_entity = [], defaultdict(list)
        for obj in objs:
            properties = obj.properties
            entity = obj.entity.name
            for key in properties.keys():
                if key not in header_per_entity[entity]:
                    header_per_entity[entity].append(key)

            properties["entity"] = entity
            properties["id"] = obj.internal_id
            all_properties.append(properties)

        header = ["entity", "id"]
        # Sort based on each entity properties
        header_per_entity = sorted(header_per_entity.items())
        for _, field_names in header_per_entity:
            header.extend(sorted(field_names))

        return header, all_properties

    def build_response(self, header, csv_data):
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = 'attachment; filename="vertices.csv"'
        writer = csv.DictWriter(response, fieldnames=header)
        writer.writeheader()
        writer.writerows(csv_data)
        return response

    def post(self, request, *args, **kwargs):
        ser = UUIDListSerializer(data=request.data)
        ser.is_valid(raise_exception=True)

        objs = get_nodes_and_properties(ser.validated_data["uuids"], serialize_method="verbose")
        header, csv_data = self.build_csv(objs)
        return self.build_response(header, csv_data)


class SavedGraphViewSet(GraphDbMixin, viewsets.ModelViewSet):
    serializer_class = SavedGraphSerializer

    def get_queryset(self):
        return SavedGraph.objects.filter(user=self.request.user).order_by("-created_at")

    def retrieve(self, request, pk=None):
        response = super().retrieve(request, pk=pk)
        obj = get_object_or_404(SavedGraph, user=request.user, pk=pk)
        edges_ids = obj.edges
        if edges_ids:
            nodes, edges = self.graph.get_by_many_edges(edges_ids)
            graph_data = self.serialize_graph(nodes, edges)
        else:
            graph_data = {"nodes": [], "edges": []}
        response.data["nodes"] = graph_data["nodes"]
        response.data["edges"] = graph_data["edges"]
        return response


class SavedGraphDetails(GraphDbMixin, APIView):
    def get(self, request, *args, **kwargs):
        graph_pk = kwargs.pop("pk")
        nodes, edges = [], []
        obj = get_object_or_404(SavedGraph, user=request.user, pk=graph_pk)
        edges_ids = obj.edges
        if edges_ids:
            nodes, edges = self.graph.get_by_many_edges(edges_ids)
        return Response(self.serialize_graph(nodes, edges))


class GraphNodeDetailEndpoint(GraphDbMixin, APIView):
    serializer_class = DetailedNodeSerializer

    def get_relationships_kwargs(self, **kwargs):
        return kwargs

    def get_serializer_class(self):
        return self.serializer_class

    def get(self, request, uuid):
        # TODO: Remove coupling between uuid variable name and url definition
        # reference: https://github.com/django/django/blob/master/django/views/generic/detail.py#L16
        obj = ObjectRepository.objects.filter(uuid=uuid).first()
        if obj is None:
            raise Http404
        data = self.get_serializer_class()(obj).data
        nodes, edges = self.graph.get_relationships(**self.get_relationships_kwargs(uuids=uuid))
        data["graph"] = self.serialize_graph(nodes, edges)
        return Response(data)


class NodeRelationshipsEndpoint(GraphDbMixin, APIView):
    def get_relationships_kwargs(self, **kwargs):
        ser = RelationshipArgumentsSerializer(data=self.request.GET)
        ser.is_valid(raise_exception=True)
        kwargs["depth"] = ser.validated_data["depth"]
        kwargs["inbound"] = ser.validated_data["inbound"]
        kwargs["outbound"] = ser.validated_data["outbound"]
        return kwargs

    def get(self, request, uuid):
        # TODO: Remove coupling between uuid variable name and url definition
        # reference: https://github.com/django/django/blob/master/django/views/generic/detail.py#L16
        # TODO: allow client to overwrite how to fetch for relationships directions

        obj = ObjectRepository.objects.filter(uuid=uuid).first()
        if obj is None:
            raise Http404
        nodes, edges = self.graph.get_relationships(**self.get_relationships_kwargs(uuids=uuid))
        if not nodes:
            nodes = [{"uuid": obj.uuid}]
        return Response(self.serialize_graphs_per_node(nodes, edges, graph_key="graph"))


class AllNodesRelationshipsEndpoint(GraphDbMixin, APIView):
    def get_relationships_kwargs(self, **kwargs):
        ser = AllNodesRelationshipsSerializer(data=self.request.GET)
        ser.is_valid(raise_exception=True)
        uuids = ser.validated_data["uuids"]
        kwargs["uuids"] = uuids
        return kwargs

    def get(self, request):
        # TODO all endpoints calling self.graph.get_relationships shoulld implement self.get_relatinoship_kwargs method
        nodes, edges = self.graph.get_relationships(**self.get_relationships_kwargs())

        return Response(self.serialize_graphs_per_node(nodes, edges, graph_key="graph"))


class GraphEdgeDetailEndpoint(GraphDbMixin, APIView):
    serializer_class = DetailedEdgeSerializer

    def get_serializer_class(self):
        return self.serializer_class

    def get(self, request, edge_id):
        # TODO: Remove coupling between edge_id variable name and url definition
        # reference: https://github.com/django/django/blob/master/django/views/generic/detail.py#L16
        try:
            nodes, edges = self.graph.get_by_edge(edge_id)
        except InvalidEdgeIdException:
            return Response({"error": f'"{edge_id}" with invalid edge ID format'}, status=400)
        if not edges:
            raise Http404

        data = self.get_serializer_class()(instance=edges[0]).data
        data["graph"] = self.serialize_graph(nodes, edges)
        return Response(data)


class SearchOnGraphEndpoint(APIView):
    search_entities = all
    search_max_items_per_entity = 10

    def get_search_term(self):
        term = self.request.GET.get("term", "").strip()
        return term

    def serialize_search_result(self, search_result):
        # Organize unique object UUIDs without losing search ranking
        unique_uuids, result_objects = [], {}
        for item in search_result:
            if item.uuid not in unique_uuids:
                unique_uuids.append(item.uuid)
                result_objects[item.uuid] = item

        # Serialize results using optimized function to group instances of same object
        nodes = {obj.uuid: obj for obj in get_nodes_and_properties(uuids=unique_uuids, serialize_method="both")}
        result = []
        for object_uuid in unique_uuids:
            obj = result_objects[object_uuid]
            serialized = nodes[object_uuid]
            result.append(
                {
                    "id": object_uuid,
                    "group": obj.entity.name,
                    "label": obj._create_label(serialized.raw_properties),
                    "properties": serialized.properties,
                }
            )
        return result

    def get(self, *args, **kwargs):
        term = self.get_search_term()
        if not term:
            return Response([])

        search_result = ObjectRepository.objects.search_many_entities(
            term,
            entities=self.search_entities,
            limit_per_entity=self.search_max_items_per_entity,
        )
        serialized_objects = self.serialize_search_result(search_result)
        return Response(serialized_objects)


class ShortestPathEndpoint(GraphDbMixin, APIView):
    def get(self, request, from_uuid, to_uuid, *args, **kwargs):
        nodes, edges = self.graph.get_shortest_path(from_uuid, to_uuid)
        return Response(self.serialize_graph(nodes, edges))


class DatasetFilter(APIView):
    def get_model(self, slug):
        dataset_models = DatasetModel.subclasses()
        if slug not in dataset_models:
            raise Http404
        return dataset_models[slug]

    def get(self, request, slug, object_uuid):
        Model = self.get_model(slug)
        if hasattr(Model.objects, "for_object"):
            objects = Model.objects.for_object(object_uuid=object_uuid)
        else:
            objects = Model.objects.filter(object_uuid=object_uuid)
        data = {
            "fields": Model.fields(),
            "result": [obj.serialize() for obj in objects],
        }
        data.update(Model.extra(objects))
        return Response(data)



import datetime
import inspect
import re
import uuid
from collections import defaultdict
from functools import lru_cache
from itertools import groupby
from textwrap import dedent
from urllib.parse import urljoin

import django.db.models.options as options
from cached_property import cached_property
from django.conf import settings
from django.contrib.postgres.aggregates import ArrayAgg
from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchQuery, SearchRank, SearchVectorField
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist
from django.db import connections, models
from django.template import Context, Template

from urlid_graph import settings as urlid_graph_settings
from urlid_graph.utils import random_name

options.DEFAULT_NAMES = options.DEFAULT_NAMES + ("entity_name", "entity_uuid", "search_fields", "slug")


REGEXP_CAMEL_CASE = re.compile("([A-Z])")


def get_urlid_database_uri():
    db = settings.DATABASES[urlid_graph_settings.DJANGO_DATABASE]
    return f"postgres://{db['USER']}:{db['PASSWORD']}@{db['HOST']}:{db['PORT']}/{db['NAME']}"


def slug_from_model_name(name):
    """
    >>> slug_from_model_name('SomeDataset')
    'some-dataset'
    >>> slug_from_model_name('MyTestModel')
    'my-test'
    """

    name = REGEXP_CAMEL_CASE.sub(r"-\1", name).lower()
    if name[0] == "-":
        name = name[1:]
    if name[-1] == "-":
        name = name[:-1]
    if name.endswith("-model"):
        name = name[:-6]
    return name


@lru_cache(maxsize=64)
def get_template(template_code):
    return Template(template_code)


class EntityQuerySet(models.QuerySet):
    def by_name(self, names):
        return self.filter(name__in=[name.lower().strip() for name in names])


class Entity(models.Model):
    objects = EntityQuerySet.as_manager()

    uuid = models.UUIDField(primary_key=True)
    base_url = models.TextField(blank=False, null=False)
    name = models.TextField(blank=False, null=False)
    version = models.TextField(blank=False, null=False)

    class Meta:
        verbose_name = "Entity"
        verbose_name_plural = "Entities"

    @property
    def url(self):
        return urljoin(self.base_url, f"/{self.name}/v{self.version}/")

    @property
    def label_properties(self):
        return getattr(self.config, "label_properties", None) or []

    @property
    def label_template(self):
        return getattr(self.config, "label_template", None) or None

    @property
    def graph_node_conf(self):
        return getattr(self.config, "graph_node_conf", None) or {}

    @property
    def config(self):
        try:
            return EntityConfig.objects.for_entity(self.name)
        except ObjectDoesNotExist:
            return None

    @property
    def label(self):
        return getattr(self.config, "label", None) or self.name

    def __str__(self):
        return f"Entity {self.url}"

    def get_model(self):
        for Model in ObjectModelMixin.__subclasses__():
            if str(Model._meta.entity_uuid) == str(self.uuid):
                return Model


class ObjectMixin:
    """Mixin used by Object and ObjectRepository classes"""

    @property
    def url(self):
        # TODO: change
        return urljoin(self.entity.url, f"{self.uuid}/")

    def _get_objects(self):
        ObjectModel = self.entity.get_model()
        if ObjectModel is None:
            raise ValueError("Model class not found fo entity UUID {}".format(self.entity.uuid))

        return ObjectModel.objects.filter(object_uuid=self.uuid).order_by(models.F("updated_at").asc(nulls_first=True))

    @cached_property
    def properties(self):
        data = {}
        for obj in self._get_objects():
            data.update(obj.serialize())
        return data

    @cached_property
    def raw_properties(self):
        data = {}
        for obj in self._get_objects():
            data.update(obj.raw_serialize())
        return data

    @cached_property
    def full_properties(self):
        data = defaultdict(list)
        for obj in self._get_objects():
            for key, value in obj.serialize():
                data[key].append(
                    {
                        "value": value,
                        "value_type": 1,  # TODO: change?
                        "source": "?",  # TODO: change?
                        "value_datetime": obj.updated_at.isoformat() if obj.updated_at else None,
                    }
                )
        return data

    def _create_label(self, raw_properties):
        label_template = self.entity.label_template
        if label_template is not None:
            return get_template(label_template).render(Context(raw_properties))
        else:
            for prop_name in self.entity.label_properties:
                label = raw_properties.get(prop_name)
                if label:
                    return label
        return f"{self.entity.name}: {self.uuid}"

    @property
    def label(self):
        return self._create_label(raw_properties=self.raw_properties)

    def get_label_for_property(self, prop_name):
        key = f"urlid_graph:ElementConfig:{self.entity.name}:{prop_name}"
        cached = cache.get(key)
        if cached is not None:
            return cached
        else:
            try:
                config = EntityPropertyConfig.objects.for_object(self.entity.name, prop_name)
                value = config.label
            except ObjectDoesNotExist:
                value = prop_name
            cache.set(key, value)
            return value

    def __str__(self):
        return f"Object {self.uuid}"


class PropertyQuerySet(models.QuerySet):
    def for_object(self, obj):
        return self.filter(object_id=obj.uuid)

    def for_objects(self, objs):
        return self.filter(object_id__in=[obj.uuid for obj in objs])


class EntityConfigManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).filter(config_type=ElementConfig.ENTITY_CONFIG)

    @lru_cache()
    def for_entity(self, entity_name):
        return self.get_queryset().get(name=entity_name)


class RelationshipConfigManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).filter(config_type=ElementConfig.REL_CONFIG)

    @lru_cache()
    def get_by_name(self, relationship_name):
        return self.get_queryset().get(name=relationship_name)


class EntityPropertyConfigManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)
        return qs.filter(config_type=ElementConfig.PROPERTY_CONFIG, parent_type=ElementConfig.ENTITY_CONFIG)

    @lru_cache()
    def for_object(self, parent_name, prop_name):
        return self.get_queryset().get(parent_name=parent_name, name=prop_name)


class RelPropertyConfigManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)
        return qs.filter(config_type=ElementConfig.PROPERTY_CONFIG, parent_type=ElementConfig.REL_CONFIG)


class StatsConfigManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)
        return qs.filter(config_type=ElementConfig.STATS_CONFIG)

    def get_stats(self):
        return self.get(name="stats")


class ElementConfig(models.Model):
    ENTITY_CONFIG, REL_CONFIG, PROPERTY_CONFIG, STATS_CONFIG, CUSTOM_CONFIG = 1, 2, 3, 4, 5
    TYPE_CHOICES = [
        (ENTITY_CONFIG, "entity"),
        (REL_CONFIG, "relationship"),
        (PROPERTY_CONFIG, "property"),
        (STATS_CONFIG, "stats"),
        (CUSTOM_CONFIG, "custom"),
    ]
    PARENT_TYPE_CHOICES = [
        (ENTITY_CONFIG, "entity"),
        (REL_CONFIG, "relationship"),
    ]

    def __str__(self):
        text = f"Config: {self.get_config_type_display()}.{self.name}"
        if self.parent_type:
            text += f" ({self.get_parent_type_display()}.{self.parent_name})"
        return text

    config_type = models.PositiveSmallIntegerField(choices=TYPE_CHOICES)
    name = models.TextField()
    parent_type = models.PositiveSmallIntegerField(choices=PARENT_TYPE_CHOICES, null=True)
    parent_name = models.TextField(null=True)
    label = models.TextField()
    data = models.JSONField()

    class Meta:
        unique_together = ["config_type", "name", "parent_type", "parent_name"]
        verbose_name = "Element configuration"
        verbose_name_plural = "Element configurations"


class EntityConfig(ElementConfig):
    objects = EntityConfigManager()

    class Meta:
        proxy = True

    @property
    def label_properties(self):
        return self.data.get("label_properties", [])

    @property
    def label_template(self):
        return self.data.get("label_template", None)

    @property
    def graph_node_conf(self):
        return self.data.get("graph_node_conf", {})


class RelationshipConfig(ElementConfig):
    objects = RelationshipConfigManager()

    class Meta:
        proxy = True

    @property
    def from_graph_node_conf(self):
        return self.data.get("from_graph_node_conf", {})

    @property
    def to_graph_node_conf(self):
        return self.data.get("to_graph_node_conf", {})


class EntityPropertyConfig(ElementConfig):
    objects = EntityPropertyConfigManager()

    class Meta:
        proxy = True


class RelPropertyConfig(ElementConfig):
    objects = RelPropertyConfigManager()

    class Meta:
        proxy = True


class StatsConfig(ElementConfig):
    objects = StatsConfigManager()

    class Meta:
        proxy = True


class ObjectRepositoryManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).defer("search_data")


class ObjectRepositoryQuerySet(models.QuerySet):
    def from_uuids(self, uuids):
        return self.filter(uuid__in=uuids).select_related("entity")

    def index(self, Model, start_id=None, lang="pg_catalog.portuguese", batch_size=10000, callback=None):
        """Update the FTS index using objects from `Model`"""

        table_name = Model._meta.db_table
        temp_table_name = "urlid_tmp_fts_" + random_name(16)
        create_temp_table_sqls = [
            (
                f'CREATE TEMPORARY TABLE "{temp_table_name}" ("id" SERIAL, "uuid" UUID)',
                [],
            ),
            (
                dedent(
                    f"""
                    INSERT INTO "{temp_table_name}" ("uuid")
                    SELECT DISTINCT object_uuid FROM "{table_name}" WHERE id >= %s
                    """
                ).strip(),
                [start_id or 0],
            ),
            (
                f'CREATE INDEX "idx_{random_name(20)}" ON "{temp_table_name}" ("id", "uuid")',
                [],
            ),
        ]
        row_count = []
        with connections[urlid_graph_settings.DJANGO_DATABASE].cursor() as cursor:
            for sql, args in create_temp_table_sqls:
                cursor.execute(sql, args)
                row_count.append(cursor.rowcount)
        total_objects = row_count[1]

        string_agg = " || ' ' || ".join(
            f"COALESCE(STRING_AGG({field_name}, ' '), '')" for field_name in Model._meta.search_fields
        )
        insert_sql = dedent(
            f"""
            INSERT INTO "{ObjectRepository._meta.db_table}" (uuid, entity_id, search_data)
                SELECT
                    object_uuid,
                    %s::uuid,
                    to_tsvector(
                        %s,
                        {string_agg}
                    )
                FROM "{table_name}"
                WHERE object_uuid IN (
                    SELECT "uuid"
                    FROM "{temp_table_name}"
                    WHERE id >= %s AND id < %s
                )
                GROUP BY object_uuid
            ON CONFLICT (uuid) DO UPDATE SET search_data = EXCLUDED.search_data
            """
        ).strip()

        counter = 0
        for object_id in range(1, total_objects + 1, batch_size):
            args = (Model._meta.entity_uuid, lang, object_id, object_id + batch_size)
            with connections[urlid_graph_settings.DJANGO_DATABASE].cursor() as cursor:
                cursor.execute(insert_sql, args)
                counter += cursor.rowcount
                if callback is not None:
                    callback(done=counter, total=total_objects)

        drop_temp_table_sql = f'DROP TABLE "{temp_table_name}"'
        with connections[urlid_graph_settings.DJANGO_DATABASE].cursor() as cursor:
            cursor.execute(drop_temp_table_sql)

        return counter

    def order_by(self, *args, **kwargs):
        # We must force search_rank ordering, since some operations will add
        # ordering after calling `.search` (like in Django Admin)
        qs = super().order_by(*args, **kwargs)
        if "-search_rank" in qs.query.order_by and qs.query.order_by[0] != "-search_rank":
            qs.query.order_by = tuple(["-search_rank"] + [item for item in qs.query.order_by if item != "-search_rank"])
        return qs

    def search(self, term, config=urlid_graph_settings.SEARCH_LANGUAGE):
        """Full-text search in object's search_data"""
        qs = self
        term = str(term or "").strip()
        if term:
            query = SearchQuery(term, config=config, search_type="websearch")
            qs = qs.annotate(search_rank=SearchRank(models.F("search_data"), query)).filter(search_data=query)
            # Overwriting `qs.query.order_by` to APPEND ordering field
            # instead of OVERWRITTING (as in `qs.order_by`). We append directly
            # (instead of using `qs.query.add_ordering` because the search rank
            # must have precedence over the other ordering.
            qs.query.order_by = tuple(["-search_rank"] + list(qs.query.order_by))
        return qs

    def search_many_entities(self, query, entities=all, limit_per_entity=10):
        """Search objects in more than one entity, by name

        Prefer to use this method since it also filter objects by entity (more performance)"""

        if entities is all:
            entities = Entity.objects.all()
        else:
            entities = Entity.objects.by_name(entities)

        results = []
        for entity in entities:
            qs = self.filter(entity=entity).search(query).select_related("entity")
            results.extend(qs[:limit_per_entity])
        results.sort(key=lambda row: row.search_rank, reverse=True)
        return results


class ObjectRepository(ObjectMixin, models.Model):
    objects = ObjectRepositoryManager.from_queryset(ObjectRepositoryQuerySet)()

    uuid = models.UUIDField(primary_key=True)
    entity = models.ForeignKey(Entity, on_delete=models.DO_NOTHING, db_constraint=False, db_index=False)
    search_data = SearchVectorField(null=True)

    class Meta:
        indexes = [GinIndex(fields=["search_data"])]
        verbose_name = "Object repository"
        verbose_name_plural = "Object repositories"


class SavedGraph(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    name = models.CharField(max_length=255)
    edges = ArrayField(models.CharField(max_length=255))
    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return f"{self.user.username} - {self.name}"

    class Meta:
        unique_together = (("user", "name"),)
        verbose_name = "Saved graph"
        verbose_name_plural = "Saved graphs"


class ObjectModelMixin(models.Model):
    object_uuid = models.UUIDField(null=False, blank=False, db_index=True)
    updated_at = models.DateTimeField(null=True, blank=True)

    @classmethod
    def serialize_object(cls, instances, related_objects, method="verbose"):
        """Group instances of the same object and convert a final version to `dict`

        Since the model can have many instances for the same `object_uuid` with different data (captured in different
        periods of time), this method group them together using `updated_at` (newer instances have precedence over
        older) and `related_objects`.
        - `method` is the final representation:
          - If equals to `"verbose"`, then model's field verbose names are used
          - If equals to `"raw"`, the model's field names are used
        """

        object_order = lambda obj: (obj.object_uuid, obj.updated_at.isoformat() if obj.updated_at else "0000-00-00")
        instances = sorted(instances, key=object_order)
        if not instances:
            return {}

        # TODO: we may not override newer values if they are `None` (this way we "loose" old but filled information)
        data = {}
        if method == "raw":
            has_related = "related_objects" in inspect.getfullargspec(type(instances[0]).raw_serialize).args
            for instance in instances:
                serialized = instance.raw_serialize(related_objects=related_objects) if has_related else instance.raw_serialize()
                data.update(serialized)
        elif method == "verbose":
            has_related = "related_objects" in inspect.getfullargspec(type(instances[0]).serialize).args
            for instance in instances:
                serialized = instance.serialize(related_objects=related_objects) if has_related else instance.serialize()
                data.update(serialized)
        return data

    def raw_serialize(self, related_objects=None, *args, **kwargs):
        """Convert an instance of an object to `dict`"""
        return {field.name: getattr(self, field.name) for field in self._meta.fields}

    def serialize(self, related_objects=None, *args, **kwargs):
        """Convert an instance of an object to `dict` using field's verbose name"""
        return {field.verbose_name: getattr(self, field.name) for field in self._meta.fields}

    class Meta:
        abstract = True


class DatasetModel(models.Model):
    object_uuid = models.UUIDField(
        verbose_name="Object UUID",
        help_text="UUID of described object",
        null=False,
        blank=False,
        db_index=True,
    )

    @classmethod
    def extra(cls, qs):
        return {}

    @classmethod
    def fields(cls):
        from django.db import connection

        return [
            {
                "name": field.name,
                "title": field.verbose_name,
                "type": field.db_type(connection),
                "description": field.help_text,
            }
            for field in cls._meta.fields
            if field.name not in ("id", "object_uuid")
        ]

    @cached_property
    def serializers(self):
        return {
            attribute[len("serialize_") :]: getattr(self, attribute)
            for attribute in dir(self)
            if attribute.startswith("serialize_")
        }

    def serialize(self):
        row = []
        for field in type(self).fields():
            field_name = field["name"]
            value = getattr(self, field["name"])
            if field_name in self.serializers:
                value = self.serializers[field_name](value)
            row.append(value)
        return row

    @classmethod
    def subclasses(cls):
        data = {}
        for klass in cls.__subclasses__():
            if hasattr(klass._meta, "slug"):
                slug = klass._meta.slug
            else:
                slug = slug_from_model_name(klass.__name__)
            data[slug] = klass
        return data

    class Meta:
        abstract = True


class JobLog(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255)
    description = models.CharField(max_length=255)
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ("-created_at",)

    @property
    def current_step(self):
        return self.steps.first().step

    @property
    def progress(self):
        step = self.steps.first()
        if step.done is not None and step.total is not None and step.total != 0:
            return f"{100 * step.done / step.total:5.2f}%"
        else:
            result = step.step + (f" ({step.action})" if step.action else "")
            if self.steps.filter(action="error").count() > 0:
                result += " - has error"
            return result

    @property
    def eta(self):
        step = self.steps.first()
        if step.done is None or step.total is None or step.total == 0:
            return "-"
        first_step = self.steps.exclude(pk=step.pk).filter(step=step.step, done__gt=0).order_by("created_at").first()
        if first_step is None:
            return "-"
        to_finish = step.total - step.done
        delta_done = step.done - first_step.done
        delta_time = (step.created_at - first_step.created_at).total_seconds()
        speed = delta_done / delta_time
        result = str(datetime.timedelta(seconds=to_finish / speed))
        return result.split(".")[0]

    @property
    def last_updated_at(self):
        return self.steps.first().created_at

    def __str__(self):
        return f"{self.name} - {self.description}: {self.id}"

    def save(self, *args, **kwargs):
        if self._state.adding:
            super().save(*args, **kwargs)
            LogStep.objects.create(job=self, action="started", step="created")
        else:
            super().save(*args, **kwargs)


class LogStep(models.Model):
    job = models.ForeignKey(JobLog, related_name="steps", on_delete=models.CASCADE)
    action = models.CharField(max_length=255)
    step = models.CharField(max_length=255)
    done = models.FloatField(null=True, blank=True)
    total = models.FloatField(null=True, blank=True)
    message = models.TextField(null=True, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ("-created_at",)

    def __str__(self):
        return f"{self.step} - {self.action}: {self.job.id}"


def get_nodes_and_properties(uuids, serialize_method="verbose"):
    """Get nodes from different entities, its related objects and serialize them in an optimized way

    - `serialize_method` is the final representation:
      - If equals to `"verbose"`, then model's field verbose names are used and saved in `properties` attribute of each
        object
      - If equals to `"raw"`, the model's field names are used and saved in `raw_properties` attribute of each object
      - If `"both"`, both are used and saved in `properties` and `raw_properties` attributes of each object
    """

    # First, detect which entities we need, so we can get all objects for each entity with one query
    objects_per_entity = {
        row["entity_id"]: row["object_uuids"]
        for row in ObjectRepository.objects.from_uuids(uuids).values("entity_id").annotate(object_uuids=ArrayAgg("uuid"))
    }
    entities = {entity.pk: entity for entity in Entity.objects.filter(uuid__in=objects_per_entity.keys())}

    # Now, serialize the objects in bulk
    verbose_serialized_objects, raw_serialized_objects = {}, {}
    for entity_id, object_uuids in objects_per_entity.items():
        ObjectModel = entities[entity_id].get_model()
        instances = ObjectModel.objects.filter(object_uuid__in=object_uuids)
        if hasattr(ObjectModel, "related_querysets"):  # Get all related objects for these `object_uuids`
            related_querysets = ObjectModel.related_querysets(object_uuids)
            related_objects_by_uuid = {}
            for obj_uuid in object_uuids:
                related_objects_by_uuid[obj_uuid] = {model_name: [] for model_name in related_querysets.keys()}
            for model_name, related_objects in related_querysets.items():
                for related_object in related_objects:
                    obj_uuid = related_object.object_uuid
                    if obj_uuid not in related_objects_by_uuid:
                        related_objects_by_uuid[obj_uuid] = {}
                    if model_name not in related_objects_by_uuid[obj_uuid]:
                        related_objects_by_uuid[obj_uuid][model_name] = []
                    related_objects_by_uuid[obj_uuid][model_name].append(related_object)
        else:
            related_objects_by_uuid = {}

        # Serialize a group of instances (same object) with the related objects
        for object_uuid, group in groupby(instances, key=lambda obj: obj.object_uuid):
            group = list(group)
            if serialize_method in ("verbose", "both"):
                verbose_serialized_objects[object_uuid] = ObjectModel.serialize_object(
                    instances=group,
                    related_objects=related_objects_by_uuid.get(object_uuid),
                    method="verbose",
                )
            if serialize_method in ("raw", "both"):
                raw_serialized_objects[object_uuid] = ObjectModel.serialize_object(
                    instances=group,
                    related_objects=related_objects_by_uuid.get(object_uuid),
                    method="raw",
                )

    # Finally, add `properties` to all `ObjectRepository` instances
    objs = list(ObjectRepository.objects.from_uuids(uuids))
    for obj in objs:
        if serialize_method in ("verbose", "both"):
            obj.__dict__["properties"] = verbose_serialized_objects.get(obj.uuid)
        if serialize_method in ("raw", "both"):
            obj.__dict__["raw_properties"] = raw_serialized_objects.get(obj.uuid)
    return objs



import uuid
from collections import namedtuple

from django.db import connections

import urlid_graph.settings as urlid_graph_settings

from .exceptions import (
    InvalidEdgeIdException,
    InvalidRelationshipsDirectionsException,
    UnexpectedCypherQueryTypeException,
)

EDGE_ID_SEPARATOR_TOKEN = "|"

EdgeInfo = namedtuple("EdgeInfo", ["label", "from_uuid", "to_uuid"])


class RelationshipsGraph:
    @property
    def db_connection(self):
        return connections[urlid_graph_settings.GRAPH_DATABASE]

    def _fetch_path(self, cursor, path_query, args={}):
        uuids_by_node_ids = {}
        cursor.execute(path_query, args)

        nodes, edges = [], {}
        for data in cursor.fetchall():
            for path_data in data:
                p_type = path_data["type"]
                if p_type != "path":
                    error_msg = f"Query must be fetching for path, not {p_type}"
                    raise UnexpectedCypherQueryTypeException(error_msg)

                for v in path_data["vertices"]:
                    if "uuid" not in v:  # Weird vertex, skip
                        # For some reason it's returning like `{'type': 'vertex', 'label': 'object', 'vid': '4.196971'}`
                        continue
                    uuids_by_node_ids[v["vid"]] = v["uuid"]
                    nodes.append({"uuid": v["uuid"]})

                for e in path_data["edges"]:
                    source_vid = e.pop("source_vid")
                    destination_vid = e.pop("destination_vid")
                    if source_vid not in uuids_by_node_ids or destination_vid not in uuids_by_node_ids:
                        continue
                    from_ = uuids_by_node_ids[source_vid]
                    to = uuids_by_node_ids[destination_vid]
                    label = e.pop("label")

                    edge_info = EdgeInfo(label, from_, to)
                    id_ = create_edge_id(edge_info)

                    edges[id_] = {
                        "from": from_,
                        "to": to,
                        "label": label,
                        "id": id_,
                        "properties": e.get("properties", {}),
                    }

        return nodes, list(edges.values())

    def get_relationships(self, uuids, depth=0, inbound=True, outbound=True):
        if isinstance(uuids, (uuid.UUID, str)):
            uuids = tuple([str(uuids)])
        elif isinstance(uuids, list):
            uuids = tuple(uuids)

        s_dir, t_dir = "-", "-"
        if inbound and not outbound:
            s_dir = "<-"
        elif outbound and not inbound:
            t_dir = "->"
        elif not inbound and not outbound:
            raise InvalidRelationshipsDirectionsException("Parameters inbound and outbound can not be both False")

        uuids_placeholders = ", ".join("%s" for _ in uuids)
        query = (
            "MATCH path=(n)<S_DIR>[*0..%s]<T_DIR>() WHERE n.uuid IN [<UUIDS>] RETURN path".replace(
                "<UUIDS>", uuids_placeholders
            )
            .replace("<S_DIR>", s_dir)
            .replace("<T_DIR>", t_dir)
        )
        nodes, edges = [], []
        query_args = (depth + 1, *uuids)
        with self.db_connection.cursor() as cursor:
            q_nodes, q_edges = self._fetch_path(cursor, query, query_args)
            nodes.extend(q_nodes)
            edges.extend(q_edges)

        return nodes, edges

    def get_by_edge(self, edge_id):
        edge_info = parse_edge_id(edge_id)

        query = (
            "MATCH path=({uuid:'<VFROM>'})-[:<ELABEL>]->({uuid:'<VTO>'}) RETURN path".replace(
                "<VFROM>", edge_info.from_uuid
            )
            .replace("<VTO>", edge_info.to_uuid)
            .replace("<ELABEL>", edge_info.label)
        )

        with self.db_connection.cursor() as cursor:
            return self._fetch_path(cursor, query)

    def get_by_many_edges(self, edges_ids):
        from_uuids, to_uuids = [], []
        for edge_id in edges_ids:
            edge_info = parse_edge_id(edge_id)
            from_uuids.append(edge_info.from_uuid)
            to_uuids.append(edge_info.to_uuid)

        query = (
            "MATCH path=(o1:object)-[]-(o2:object)"
            " WHERE o1.uuid IN {from_uuids} AND o2.uuid IN {to_uuids} RETURN path;"
        ).format(from_uuids=from_uuids, to_uuids=to_uuids)

        with self.db_connection.cursor() as cursor:
            return self._fetch_path(cursor, query)

    def get_shortest_path(self, from_uuid, to_uuid, limit=20):
        # There is also allShortestPaths
        query = (
            "MATCH (o1: object), (o2: object),"
            f" path=shortestpath((o1)-[*..{limit}]-(o2))"
            f" WHERE o1.uuid = '{from_uuid}' AND o2.uuid = '{to_uuid}'"
            " RETURN path;"
        )
        with self.db_connection.cursor() as cursor:
            return self._fetch_path(cursor, query)


def parse_edge_id(value):
    data = value.split(EDGE_ID_SEPARATOR_TOKEN)
    if len(data) != 3 or not data[0]:
        raise InvalidEdgeIdException(f"Value {repr(value)} not in format 'label|from_uuid|to_uuid'")

    return EdgeInfo(*data)


def create_edge_id(edge_info):
    return EDGE_ID_SEPARATOR_TOKEN.join(edge_info)


def get_relationships_graph():
    return RelationshipsGraph()



from pathlib import Path

from django.apps import apps
from django.core.management import call_command
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Import objects and relationships from a directory"

    def add_arguments(self, parser):
        parser.add_argument("path")
        parser.add_argument("--disable-autovacuum", action="store_true")
        parser.add_argument("--disable-indexes", action="store_true")
        parser.add_argument("--filetype", action="append")

    def handle(self, *args, **options):
        cmd_options = {
            "disable_autovacuum": options["disable_autovacuum"],
            "disable_indexes": options["disable_indexes"],
        }
        cmd_options_relations = {
            "disable_autovacuum": options["disable_autovacuum"],
        }
        path = Path(options["path"])
        file_types = options["filetype"]
        if not file_types:
            file_types = all
        else:
            file_types = [file_type.lower() for file_type in file_types]
            for file_type in file_types:
                if file_type not in ("object", "relationship"):
                    raise ValueError(f"Unknown value '{file_type}'")

        # Import files containing objects
        # TODO: should use ObjectModelMixin.__subclasses__() so we don't need
        # to traverse all apps/models.
        model_app = {
            model_name.lower(): app_name
            for app_name, app_models in apps.all_models.items()
            for model_name, Model in app_models.items()
        }
        if file_types is all or "object" in file_types:
            filenames = [str(filename.absolute()) for filename in path.glob("*object*.csv*")]
            oks = []
            for filename in filenames:
                model = filename.split("object")[1].split(".csv")[0].replace("-", "").replace("_", "")
                app_name = model_app[model]
                oks.append(call_command("import_objects", app_name, model, filename, **cmd_options) == "True")
            ok = all(oks)
            if not ok:
                return str(ok)

        # Import files containing relationships
        if file_types is all or "relationship" in file_types:
            oks = []
            for filename in path.glob("*relationship*.csv*"):
                relationship = filename.absolute().name.split("relationship")[1].split(".csv")[0]
                filename = str(filename.absolute())
                if relationship.startswith("-") or relationship.startswith("_"):
                    relationship = relationship[1:]
                if relationship.endswith("-") or relationship.endswith("_"):
                    relationship = relationship[:-1]

                oks.append(
                    call_command("import_relationships", relationship, filename, **cmd_options_relations) == "True"
                )
            ok = all(oks)
            if not ok:
                return str(ok)

        return "True"



import traceback
from pathlib import Path

import rows
from django.apps import apps
from django.core.management.base import BaseCommand
from django.db import connections

import urlid_graph.settings as urlid_graph_settings
from urlid_graph.log import ImportObjectsLogger
from urlid_graph.models import DatasetModel, ObjectModelMixin, ObjectRepository, get_urlid_database_uri
from urlid_graph.utils import DatabaseConnection, read_total_size


class Command(BaseCommand):
    help = "Import objects into the correct model"

    def add_arguments(self, parser):
        parser.add_argument("--batch-size", type=int, default=100_000, help="Number of rows to index per turn")
        parser.add_argument(
            "--chunk-size", type=int, default=8_388_608, help="Chunk size used to read CSV file while importing data"
        )
        parser.add_argument("--disable-autovacuum", action="store_true")
        parser.add_argument("--disable-indexes", action="store_true")
        parser.add_argument("app_name")
        parser.add_argument("model")
        parser.add_argument("input_filename")

    def handle(self, *args, **options):
        app_name = options["app_name"]
        model = options["model"]
        input_filename = options["input_filename"]
        batch_size = options["batch_size"]
        chunk_size = options["chunk_size"]
        disable_autovacuum = options["disable_autovacuum"]
        disable_indexes = options["disable_indexes"]

        Model = apps.get_model(app_name, model)
        if not issubclass(Model, (DatasetModel, ObjectModelMixin)):
            raise ValueError("Model '{}' doesn't inherit from `ObjectModelMixin` or `DatasetModel`".format(model))
        table_name = Model._meta.db_table
        connection = connections[urlid_graph_settings.DJANGO_DATABASE]
        db = DatabaseConnection(connection=connection)

        self.logger = ImportObjectsLogger(description=f"{Path(input_filename).name} -> {app_name}.{model}")
        print(f"Starting job {self.logger.job.id}")

        ok = True
        try:
            with self.logger.step("pre-import", message_start="Preparing table to import"):
                self.logger.message("Disabling sync commit")
                db.disable_sync_commit()

                if disable_autovacuum:
                    self.logger.message(f"Disabling autovacuum for {table_name}")
                    db.disable_autovacuum(table_name)

                if disable_indexes:
                    self.logger.message(f"Disabling table indexes for {table_name}")
                    db.execute_query(
                        f"UPDATE pg_index SET indisready = FALSE WHERE indrelid = '{table_name}'::regclass"
                    )

                self.logger.message(f"Disabling triggers for {table_name}")
                db.execute_query(f'ALTER TABLE "{table_name}" DISABLE TRIGGER ALL')

            last_object = Model.objects.order_by("-id").first()
            start_id = last_object.id + 1 if last_object is not None else None
            self.import_data(input_filename, chunk_size, table_name)

            if issubclass(Model, ObjectModelMixin):
                # Don't need to add to search index if it's just a custom
                # dataset
                self.add_to_search_index(Model, start_id, batch_size)

        except:  # noqa
            self.logger.error(traceback.format_exc())
            ok = False

        finally:
            with self.logger.step("post-import"):
                if disable_indexes:
                    self.logger.message(f"Reindexing table {table_name}")
                    db.execute_query(f'REINDEX TABLE "{table_name}"')

                self.logger.message(f"Reenabling triggers on {table_name}")
                db.execute_query(f'ALTER TABLE "{table_name}" ENABLE TRIGGER ALL')

                self.logger.message(f"Running VACUUM ANALYZE on {table_name}")
                db.vacuum_analyze(table_name)

                if disable_autovacuum:
                    self.logger.message(f"Enabling autovacuum for {table_name}")
                    db.enable_autovacuum(table_name)

                self.logger.message("Enabling sync commit")
                db.enable_sync_commit()

        self.logger.finish()
        return str(ok)  # Used by import_data when calling this command programatically

    def import_data(self, input_filename, chunk_size, table_name):
        self.logger.start_step("import")
        file_size = read_total_size(input_filename)

        def update_progress(done, total):
            # done is actually the last step done and total the total_done
            self.logger.progress(done=total, total=file_size)

        try:
            rows_output = rows.utils.pgimport(
                input_filename,
                encoding="utf-8",
                dialect="excel",
                table_name=table_name,
                database_uri=get_urlid_database_uri(),
                create_table=False,
                chunk_size=chunk_size,
                callback=update_progress,
            )
        except RuntimeError:
            self.logger.error(traceback.format_exc())
            raise
        else:
            self.logger.finish_step("import", message=f"Imported {rows_output['rows_imported']} rows to '{table_name}'")

    def add_to_search_index(self, Model, start_id, batch_size):
        self.logger.start_step("index", message=f"start_id: {start_id}, batch_size: {batch_size}")

        ObjectRepository.objects.index(
            Model,
            start_id=start_id,
            batch_size=batch_size,
            callback=self.logger.progress,
        )
        self.logger.finish_step("index")



from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import DataError
from django.db.backends.signals import connection_created

from urlid_graph import settings as urlid_graph_settings


def set_graph_path(sender, connection, **kwargs):
    graph_name = urlid_graph_settings.GRAPH_NAME
    if graph_name and connection.alias == urlid_graph_settings.GRAPH_DATABASE:
        with connection.cursor() as cursor:
            try:
                cursor.execute(f"SET graph_path = {graph_name}")
            except DataError:
                cursor.execute(f"CREATE graph {graph_name}")
                cursor.execute(f"SET graph_path = {graph_name}")


class DataGraphConfig(AppConfig):
    name = "urlid_graph"

    def ready(self):
        if urlid_graph_settings.TESTING_ENV:
            # TESTING_ENV is used so projects which depends on this app can run
            # the tests without configuring everything needed by this app (like
            # a graph database).
            return

        urlid_graph_db_router = "urlid_graph.db_router.RelationAndGraphDBRouter"
        if "rest_framework" not in settings.INSTALLED_APPS:
            raise ImproperlyConfigured("rest_framework must be in installed apps.")
        if getattr(settings, "GRAPH_DATABASE", None) is None:
            raise ImproperlyConfigured("must set a value for GRAPH_DATABASE in settings.py")
        if getattr(settings, "GRAPH_DATABASE_URL", None) is None:
            raise ImproperlyConfigured("must set a value for GRAPH_DATABASE_URL in settings.py")
        if settings.GRAPH_DATABASE not in settings.DATABASES:
            raise ImproperlyConfigured("must set GRAPH_DATABASE with GRAPH_DATABASE_URL in DATABASES")
        if urlid_graph_db_router not in settings.DATABASE_ROUTERS:
            raise ImproperlyConfigured(f"must set {urlid_graph_db_router} in DATABASE_ROUTERS")

        connection_created.connect(set_graph_path)



from urlid_graph import settings as urlid_graph_settings

EXPANDED_SUFFIX = "_expanded"


graph_vis_options = {
    "all_nodes_relationships_chunk_size": urlid_graph_settings.ALL_NODES_RELATIONSHIPS_CHUNK_SIZE,
    "expanded_node_group_suffix": EXPANDED_SUFFIX,
    "edges": {
        "arrows": {"to": {"enabled": True}},
        "color": {"color": "#ed8936", "highlight": "#ed8936"},
        "font": {"color": "#1a202c", "align": "top"},
    },
    "nodes": {
        "font": {"color": "#1a202c", "strokeWidth": 2, "strokeColor": "#ffffff"},
    },
    "physics": {
        "enabled": True,
        "barnesHut": {"gravitationalConstant": -50000, "springConstant": 0.01},
        "stabilization": {"enabled": True, "iterations": 2000, "updateInterval": 10},
    },
    "groups": {},
}


def get_entity_node_config(entity):
    node_cfg = entity.graph_node_conf or {}
    background_color_collapsed = node_cfg.get("background-collapsed", "#61b2fe")
    background_color_expanded = node_cfg.get("background-expanded", "#3182ce")
    color_collapsed = node_cfg.get("color-collapsed", "#fff")
    color_expanded = node_cfg.get("color-expanded", "#fff")
    icon = icon_expanded = node_cfg.get("code", "")

    return {
        entity.name: {
            "color": {
                "border": background_color_collapsed,
                "background": background_color_collapsed,
                "highlight": {
                    "border": "white",
                    "background": background_color_collapsed,
                },
            },
            "shape": "dot",
            "icon": {
                "code": icon,
                "color": color_collapsed,
                "description": f"{entity.label} sem relações expandidas",
                "face": '"Font Awesome 5 Free"',
                "size": 50,
            },
        },
        f"{entity.name}{EXPANDED_SUFFIX}": {
            "shadow": {"enabled": True, "size": "8"},
            "color": {
                "border": background_color_expanded,
                "background": background_color_expanded,
                "highlight": {
                    "border": "white",
                    "background": background_color_expanded,
                },
            },
            "shape": "dot",
            "icon": {
                "code": icon_expanded,
                "color": color_expanded,
                "description": f"{entity.label} com todas as relações já expandidas",
                "face": '"Font Awesome 5 Free"',
                "size": 50,
            },
        },
    }



from cached_property import cached_property

from .graph_db import get_relationships_graph
from .models import get_nodes_and_properties
from .serializers import DetailedEdgeSerializer, DetailedNodeSerializer


class GraphDbMixin:
    node_serializer_class = DetailedNodeSerializer
    edge_serializer_class = DetailedEdgeSerializer

    @cached_property
    def graph(self):
        return get_relationships_graph()

    def serialize_graph(self, nodes, edges):
        """
        nodes and edges as returned by RelationshipsGraph query methods
        """
        nodes_objs = get_nodes_and_properties(uuids=[n["uuid"] for n in nodes], serialize_method="both")
        return self.serialize_graph_objects(nodes_objs, edges)

    def serialize_graph_objects(self, nodes_objs, edges):
        """
        nodes_objs are Object and edges are edges returned by RelationshipsGraph
        """
        return {"nodes": self.serialize_nodes(nodes_objs), "edges": self.serialize_edges(edges)}

    def serialize_graphs_per_node(self, nodes, edges, graph_key, serializer_class=None):
        serializer_class = serializer_class or self.get_node_serializer_class()
        data = {}
        nodes_objs = get_nodes_and_properties(uuids=[n["uuid"] for n in nodes], serialize_method="both")
        for node in nodes_objs:
            node_uuid = str(node.uuid)
            graph_edges = [e for e in edges if node_uuid in (e["to"], e["from"])]
            graph_nodes = set()
            for edge in graph_edges:
                from_node = [n for n in nodes_objs if str(n.uuid) == edge["from"]]
                to_node = [n for n in nodes_objs if str(n.uuid) == edge["to"]]
                if from_node:
                    graph_nodes.add(from_node[0])
                if to_node:
                    graph_nodes.add(to_node[0])
            node_data = serializer_class(node).data
            node_data[graph_key] = self.serialize_graph_objects(graph_nodes, graph_edges)
            data[node_uuid] = node_data

        return data

    def serialize_nodes(self, nodes_objs):
        Serializer = self.get_node_serializer_class()
        return Serializer(nodes_objs, many=True).data

    def serialize_edges(self, edges_objs):
        Serializer = self.get_edge_serializer_class()
        return Serializer(edges_objs, many=True).data

    def get_node_serializer_class(self):
        return self.node_serializer_class

    def get_edge_serializer_class(self):
        return self.edge_serializer_class

    def get_relationships_kwargs(self, **kwargs):
        raise NotImplementedError



import urlid_graph.settings as urlid_graph_settings


class RelationAndGraphDBRouter:
    def db_for_read(self, model, **hints):
        return "default"

    def db_for_write(self, model, **hints):
        return "default"

    def allow_relation(self, obj1, obj2, **hints):
        return True

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if "target_db" in hints:
            return db == urlid_graph_settings.GRAPH_DATABASE
        return db == "default"



# Generated by Django 4.0.6 on 2022-08-02 09:44

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('urlid_graph', '0003_joblog_jobstep'),
    ]

    operations = [
        migrations.AlterField(
            model_name='joblog',
            name='id',
            field=models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='logstep',
            name='message',
            field=models.TextField(blank=True, null=True),
        ),
    ]
